#include <zephyr/kernel.h>
#include <zephyr/device.h>
#include <zephyr/drivers/pwm.h>
#include <zephyr/sys/util.h>
#include <zephyr/logging/log.h>
#include "TaskBEEP.h"
#include "hw_init.h"
#include "main.h"

LOG_MODULE_REGISTER(task_beep, LOG_LEVEL_INF);

#define STACK_SIZE         1024
#define BEEP_PRIORITY      0
#define BEEP_MSGQ_LEN      16
#define BEEP_INFINITY_REPEAT 0xFFFFFFFF

// ======= Thread stack & message queue =======
K_THREAD_STACK_DEFINE(beep_stack_area, STACK_SIZE);
static struct k_thread beep_thread;
K_MSGQ_DEFINE(beep_msgq, sizeof(beep_msg_t), BEEP_MSGQ_LEN, 4);

static struct pwm_dt_spec beep_pwm;

// ===== Patterny a notová tabulka (vložíš si svoje, jen příklad) =====
static const uint16_t note_freq_table[NOTE_COUNT] = {
    [NOTE_SILENT] = 0,
    [NOTE_C4] = 261,  [NOTE_CS4] = 277,  [NOTE_D4] = 293,  [NOTE_DS4] = 311,  [NOTE_E4] = 329,  [NOTE_F4] = 349,  [NOTE_FS4] = 370,
    [NOTE_G4] = 392,  [NOTE_GS4] = 415,  [NOTE_A4] = 440,  [NOTE_AS4] = 466,  [NOTE_B4] = 493,
    [NOTE_C5] = 523,  [NOTE_CS5] = 554,  [NOTE_D5] = 587,  [NOTE_DS5] = 622,  [NOTE_E5] = 659,  [NOTE_F5] = 698,  [NOTE_FS5] = 740,
    [NOTE_G5] = 784,  [NOTE_GS5] = 831,  [NOTE_A5] = 880,  [NOTE_AS5] = 932,  [NOTE_B5] = 987,
    [NOTE_C6] = 1047, [NOTE_CS6] = 1109, [NOTE_D6] = 1175, [NOTE_DS6] = 1245, [NOTE_E6] = 1319, [NOTE_F6] = 1397, [NOTE_FS6] = 1480,
    [NOTE_G6] = 1568, [NOTE_GS6] = 1661, [NOTE_A6] = 1760, [NOTE_AS6] = 1865, [NOTE_B6] = 1976,
    [NOTE_C7] = 2093, [NOTE_CS7] = 2217, [NOTE_D7] = 2349, [NOTE_DS7] = 2489, [NOTE_E7] = 2637, [NOTE_F7] = 2794, [NOTE_FS7] = 2960,
    [NOTE_G7] = 3136, [NOTE_GS7] = 3322, [NOTE_A7] = 3520, [NOTE_AS7] = 3729, [NOTE_B7] = 3951
};

// ----- Pattern struktury -----
typedef struct
{
    note_t note;
    uint32_t duration_ms;
} beep_pattern_step_t;

typedef struct
{
    const beep_pattern_step_t *steps;
    uint8_t step_count;
} beep_pattern_t;

// --- Příklad patternů ---
static const beep_pattern_step_t beep_btn_pressed_pattern[] = {
    { NOTE_A5, 100 }, { NOTE_SILENT, 10 }
};
static const beep_pattern_step_t beep_btn_pressed_long_pattern[] = {
    { NOTE_A5, 500 }, { NOTE_SILENT, 10 }
};
static const beep_pattern_step_t pattern_system_on[] = {
    { NOTE_C5, 150 }, { NOTE_E5, 150 }, { NOTE_G5, 200 }
};

static const beep_pattern_step_t pattern_system_off[] = {
    { NOTE_G5, 150 }, { NOTE_E5, 150 }, { NOTE_C5, 200 }
};

// Příjemný/přátelský, střední výška tónu, delší mezery, krátké tóny
static const beep_pattern_step_t pattern_zone1[] = {
    { NOTE_E5, 120 }, { NOTE_SILENT, 180 }, { NOTE_E5, 120 }
};

// Naléhavější, vyšší tón, rychlejší opakování
static const beep_pattern_step_t pattern_zone2[] = {
    { NOTE_G5, 100 }, { NOTE_SILENT, 80 }, { NOTE_G5, 100 }, { NOTE_SILENT, 80 }, { NOTE_G5, 100 }
};

// Vysoký tón, krátké opakování, záměrně „nepříjemný“ zvuk
static const beep_pattern_step_t pattern_zone3[] = {
    { NOTE_C6, 70 }, { NOTE_SILENT, 40 }, { NOTE_DS5, 70 }, { NOTE_SILENT, 40 },
    { NOTE_C6, 70 }, { NOTE_SILENT, 40 }, { NOTE_DS5, 70 }
};
static const beep_pattern_step_t pattern_zone3_siren[] = {
    { NOTE_B5, 60 }, { NOTE_DS6, 60 }, { NOTE_B5, 60 }, { NOTE_DS6, 60 },
    { NOTE_B5, 60 }, { NOTE_DS6, 60 }, { NOTE_B5, 60 }, { NOTE_DS6, 60 }
};

static const beep_pattern_step_t pattern_zone3_rising[] = {
    { NOTE_G5, 60 }, { NOTE_A5, 60 }, { NOTE_B5, 60 }, { NOTE_C6, 60 },
    { NOTE_G5, 60 }, { NOTE_A5, 60 }, { NOTE_B5, 60 }, { NOTE_C6, 60 }
};

static const beep_pattern_step_t pattern_zone3_minor2nd[] = {
    { NOTE_C6, 40 }, { NOTE_CS6, 40 }, { NOTE_C6, 40 }, { NOTE_CS6, 40 },
    { NOTE_C6, 40 }, { NOTE_CS6, 40 }, { NOTE_C6, 40 }, { NOTE_CS6, 40 }
};

static const beep_pattern_step_t pattern_zone3_random[] = {
    { NOTE_GS5, 50 }, { NOTE_C6, 50 }, { NOTE_DS6, 50 }, { NOTE_A5, 50 },
    { NOTE_F5, 50 }, { NOTE_B5, 50 }, { NOTE_GS5, 50 }, { NOTE_C6, 50 }
};

static const beep_pattern_step_t pattern_zone3_aggressive[] = {
    { NOTE_C6, 40 }, { NOTE_FS6, 40 }, { NOTE_C6, 40 }, { NOTE_FS6, 40 },
    { NOTE_D6, 40 }, { NOTE_G6, 40 }, { NOTE_D6, 40 }, { NOTE_G6, 40 }
};




// --- Pattern array ---
static const beep_pattern_t beep_patterns[] = {
    [BEEP_CMD_OFF]        = { NULL, 0 },
    [BEEP_CMD_ON]         = { NULL, 0 },
    [BEEP_SHORT_PRESS]    = { beep_btn_pressed_pattern, ARRAY_SIZE(beep_btn_pressed_pattern) },
    [BEEP_LONG_PRESS]     = { beep_btn_pressed_long_pattern, ARRAY_SIZE(beep_btn_pressed_long_pattern) },
    [BEEP_CMD_SYSTEM_ON]  = { pattern_system_on, ARRAY_SIZE(pattern_system_on) },
    [BEEP_CMD_SYSTEM_OFF] = { pattern_system_off, ARRAY_SIZE(pattern_system_off) },
    [BEEP_TONE_ZONE1]   = { pattern_zone1, ARRAY_SIZE(pattern_zone1) },
    [BEEP_TONE_ZONE2]   = { pattern_zone2, ARRAY_SIZE(pattern_zone2) }, 
    [BEEP_TONE_ZONE3]   = { pattern_zone3, ARRAY_SIZE(pattern_zone3) },
    [BEEP_TONE_ZONE4]   = { pattern_zone3_siren, ARRAY_SIZE(pattern_zone3_siren) },
    [BEEP_TONE_ZONE5]   = { pattern_zone3_rising, ARRAY_SIZE(pattern_zone3_rising) },
    [BEEP_TONE_ZONE6]   = { pattern_zone3_minor2nd, ARRAY_SIZE(pattern_zone3_minor2nd) },
    [BEEP_TONE_ZONE7]   = { pattern_zone3_aggressive, ARRAY_SIZE(pattern_zone3_aggressive) },
    // Další patterny si přidej podle potřeby...
};

_Static_assert(BEEP_CMD_COUNT == ARRAY_SIZE(beep_patterns), "Počet patternů neodpovídá počtu enumů beep_cmd_t!");

// ======= Task context =======
typedef enum
{
    BEEP_STATE_OFF,
    BEEP_STATE_TONE,
    BEEP_STATE_PATTERN
} beep_state_t;

typedef struct
{
    const beep_pattern_t *pattern;
    uint8_t current_step;
    uint32_t repeat_counter;
    beep_state_t state;
    struct k_timer timer;
    bool has_pending_msg;
    beep_msg_t pending_msg;
} beep_ctx_t;

static beep_ctx_t beep_ctx;

// ======= Prototypy =======
static void Task_Beep(void *arg1, void *arg2, void *arg3);
static void Callback_BEEP_Timer(struct k_timer *timer);
static void beep_start_pattern(beep_ctx_t *ctx, const beep_msg_t *msg);
static void beep_output(note_t note, uint8_t volume_percent);
static const beep_pattern_t *get_beep_pattern(beep_cmd_t cmd);

// ======= Pattern getter =======
static const beep_pattern_t *get_beep_pattern(beep_cmd_t cmd)
{
    if ((uint8_t)cmd < ARRAY_SIZE(beep_patterns)) {
        return &beep_patterns[cmd];
    }
    LOG_ERR("Unknown beep cmd: %d", cmd);
    return NULL;
}


// ======= PWM output =======
static void beep_output(note_t note, uint8_t volume_percent)
{
    if ((uint8_t)note >= NOTE_COUNT)
    {
        LOG_ERR("Invalid note index: %d", note);
        return;
    }
    uint16_t freq = note_freq_table[note];
    if (freq == 0)
    {
        pwm_set_dt(&beep_pwm, PWM_MSEC(1), 0);
        return;
    }
    uint32_t period_ns = 1000000000UL / freq;
    if (volume_percent > 100) volume_percent = 100;
    // Duty cycle = hlasitost v %
    uint32_t pulse_ns = (period_ns * volume_percent) / 100;
    LOG_DBG("Beep output: note=%d, freq=%d, period_ns=%d, pulse_ns=%d", note, freq, period_ns, pulse_ns);

    int ret = pwm_set_dt(&beep_pwm, period_ns, pulse_ns);
    if (ret != 0)
    {
        LOG_ERR("Failed to set PWM: %d", ret);
    }
}


// ======= Pattern start =======
static void beep_start_pattern(beep_ctx_t *ctx, const beep_msg_t *msg)
{
    const beep_pattern_t *pat = get_beep_pattern(msg->cmd);
    ctx->pattern = pat;
    ctx->current_step = 0;
    ctx->repeat_counter = (msg->repeat_count == 0) ? 1 : msg->repeat_count;
    ctx->state = (pat != NULL) ? BEEP_STATE_PATTERN : BEEP_STATE_OFF;
    ctx->has_pending_msg = false;

    if (pat != NULL && pat->step_count > 0)
    {
        beep_output(pat->steps[0].note,BEEP_VOLUME_PERCENT);
        k_timer_start(&ctx->timer, K_MSEC(pat->steps[0].duration_ms), K_NO_WAIT);
    }
    else if (msg->cmd == BEEP_CMD_ON)
    {
        beep_output(msg->note,BEEP_VOLUME_PERCENT);
        k_timer_stop(&ctx->timer);
        ctx->state = BEEP_STATE_TONE;
    }
    else // OFF nebo neplatné
    {
        beep_output(NOTE_SILENT,BEEP_VOLUME_PERCENT);
        k_timer_stop(&ctx->timer);
        ctx->state = BEEP_STATE_OFF;
    }
}

// ======= Timer callback =======
static void Callback_BEEP_Timer(struct k_timer *timer)
{
    beep_ctx_t *ctx = CONTAINER_OF(timer, beep_ctx_t, timer);

    if (ctx->pattern == NULL)
    {
        return;
    }
    ctx->current_step++;

    if (ctx->current_step >= ctx->pattern->step_count)
    {
        ctx->current_step = 0;
        if (ctx->repeat_counter != BEEP_INFINITY_REPEAT)
        {
            if (ctx->repeat_counter > 0)
            {
                ctx->repeat_counter--;
            }
            if (ctx->repeat_counter == 0)
            {
                ctx->state = BEEP_STATE_OFF;
                beep_output(NOTE_SILENT,BEEP_VOLUME_PERCENT);
                if (ctx->has_pending_msg)
                {
                    beep_start_pattern(ctx, &ctx->pending_msg);
                    ctx->has_pending_msg = false;
                }
                return;
            }
        }
    }

    if (ctx->pattern != NULL)
    {
        beep_output(ctx->pattern->steps[ctx->current_step].note,BEEP_VOLUME_PERCENT);
        k_timer_start(&ctx->timer, K_MSEC(ctx->pattern->steps[ctx->current_step].duration_ms), K_NO_WAIT);
    }
}

// ======= API =======
void TaskBeep_Start(void)
{   
    beep_pwm = *get_pwm_beep();
    k_thread_create(&beep_thread, beep_stack_area, K_THREAD_STACK_SIZEOF(beep_stack_area),
                    Task_Beep, NULL, NULL, NULL,
                    BEEP_PRIORITY, 0, K_NO_WAIT);

    k_thread_name_set(&beep_thread, "TaskBeep");
}

int TaskBeep_SendCommand(const beep_msg_t *msg)
{
    return k_msgq_put(&beep_msgq, msg, K_FOREVER);
}

// ======= Main Task =======
static void Task_Beep(void *arg1, void *arg2, void *arg3)
{
    k_timer_init(&beep_ctx.timer, Callback_BEEP_Timer, NULL);
    beep_ctx.state = BEEP_STATE_OFF;
    beep_ctx.pattern = NULL;
    beep_ctx.current_step = 0;
    beep_ctx.repeat_counter = 0;
    beep_ctx.has_pending_msg = false;
    beep_output(NOTE_SILENT,BEEP_VOLUME_PERCENT);

    beep_msg_t msg;
    while (1)
    {   
        log_stack_usage("TaskBeep");

        k_msgq_get(&beep_msgq, &msg, K_FOREVER);

        if (msg.wait_for_previous && beep_ctx.state == BEEP_STATE_PATTERN)
        {
            beep_ctx.pending_msg = msg;
            beep_ctx.has_pending_msg = true;
        }
        else
        {
            beep_start_pattern(&beep_ctx, &msg);
        }
    }
}
