// TaskButton.c
#include "TaskButton.h"
#include <zephyr/logging/log.h>
#include "hw_init.h"
#include "TaskBeep.h"
#include "TaskLED.h"
#include "TaskMain.h"
#include "main.h"

LOG_MODULE_REGISTER(task_button, LOG_LEVEL_INF);

#define DEBOUNCE_MS 10
#define HOLD_CHECK_MS 30

#define SHORT_PRESS_MS 300
#define LONG_PRESS_MS 1200
#define EXTRA_PRESS_MS 3000
#define LONGEST_PRESS_MS 10000

static struct gpio_callback button_cb;
static struct k_work_delayable debounce_work;
static struct k_timer hold_timer;

static struct k_thread button_thread;
#define BUTTON_STACK_SIZE 1024
#define BUTTON_PRIORITY 0
K_THREAD_STACK_DEFINE(button_stack_area, BUTTON_STACK_SIZE);

#define MSGQ_SIZE 16
K_MSGQ_DEFINE(button_msgq, sizeof(button_event_t), MSGQ_SIZE, 4);

static atomic_t suppress_events;
static atomic_t is_pressed;
static uint32_t press_time_ms;
static uint32_t last_notified;
static button_event_type_t hold_type;

static const struct gpio_dt_spec *button;

static void Task_Button(void);

bool TaskButton_IsPressed(void)
{
    return atomic_get(&is_pressed);
}

int TaskButton_GetEvent(button_event_t *event, k_timeout_t timeout)
{
    return k_msgq_get(&button_msgq, event, timeout);
}

static void emit_event(button_event_type_t type)
{
    if (!atomic_get(&suppress_events))
    {
        button_event_t evt = {
            .type = type,
            .timestamp_ms = k_uptime_get_32()
        };
        if (k_msgq_put(&button_msgq, &evt, K_NO_WAIT) != 0)
        {
            LOG_WRN("Message queue full, dropping button event");
        }
    }
}

static void hold_timer_handler(struct k_timer *timer)
{
    uint32_t held = k_uptime_get_32() - press_time_ms;

    if (held >= LONGEST_PRESS_MS && last_notified < LONGEST_PRESS_MS)
    {
        emit_event(BUTTON_EVENT_LONGEST);
        last_notified = held;
        hold_type = BUTTON_EVENT_LONGEST;
    }
    else if (held >= EXTRA_PRESS_MS && last_notified < EXTRA_PRESS_MS)
    {
        emit_event(BUTTON_EVENT_EXTRA);
        last_notified = held;
        hold_type = BUTTON_EVENT_EXTRA;
    }
    else if (held >= LONG_PRESS_MS && last_notified < LONG_PRESS_MS)
    {
        emit_event(BUTTON_EVENT_LONG);
        last_notified = held;
        hold_type = BUTTON_EVENT_LONG;
    }
}

static void debounce_fn(struct k_work *work)
{
    int val = gpio_pin_get_dt(button);
    uint32_t now = k_uptime_get_32();

    bool is_active = (val == BUTTON_ACTIVE_STATE);

    LOG_INF("Button state: %s", is_active ? "pressed" : "released");

    if (is_active)
    {
        if (!atomic_get(&is_pressed))
        {
            atomic_set(&is_pressed, true);
            press_time_ms = now;
            last_notified = 0;
            hold_type = BUTTON_EVENT_SHORT;
            k_timer_start(&hold_timer, K_MSEC(HOLD_CHECK_MS), K_MSEC(HOLD_CHECK_MS));
            emit_event(BUTTON_EVENT_PRESSED);
        }
    }
    else
    {
        if (atomic_get(&is_pressed))
        {
            k_timer_stop(&hold_timer);
            atomic_set(&is_pressed, false);
            emit_event(BUTTON_EVENT_RELEASED);

            if (last_notified == 0)
            {
                emit_event(hold_type);
            }
        }
    }
}

static void button_isr(const struct device *dev, struct gpio_callback *cb, uint32_t pins)
{
    k_work_schedule(&debounce_work, K_MSEC(DEBOUNCE_MS));
}

void TaskButton_Start(void)
{
    button = get_button();

    gpio_pin_configure_dt(button, GPIO_INPUT);
    gpio_pin_interrupt_configure_dt(button, GPIO_INT_EDGE_BOTH);

    gpio_init_callback(&button_cb, button_isr, BIT(button->pin));
    gpio_add_callback(button->port, &button_cb);

    k_work_init_delayable(&debounce_work, debounce_fn);
    k_timer_init(&hold_timer, hold_timer_handler, NULL);

    atomic_clear(&suppress_events);
    atomic_clear(&is_pressed);

    k_thread_create(&button_thread, button_stack_area, BUTTON_STACK_SIZE,
                    Task_Button, NULL, NULL, NULL,
                    BUTTON_PRIORITY, 0, K_NO_WAIT);

    k_thread_name_set(&button_thread, "TaskButton");


    LOG_INF("TaskButton initialized");
}

void TaskButton_Suppress(bool suppress)
{
    atomic_set(&suppress_events, suppress);
    LOG_INF("Button events %s", suppress ? "suppressed" : "enabled");
}

static void Task_Button(void)
{
    while (1)
    {   
        log_stack_usage("TaskButton");

        button_event_t evt;
        if (TaskButton_GetEvent(&evt, K_FOREVER) == 0)
        {
            main_event_t main_evt = {0};

            switch (evt.type)
            {
                case BUTTON_EVENT_PRESSED:
                    main_evt.type = MAIN_EVENT_BTN_PRESSED_EDGE;
                    LOG_INF("Button pressed");
                    break;
                case BUTTON_EVENT_RELEASED:
                    main_evt.type = MAIN_EVENT_BTN_RELEASED;
                    LOG_INF("Button released");
                    break;
                case BUTTON_EVENT_SHORT:
                    main_evt.type = MAIN_EVENT_BTN_SHORT_PRESS;
                    LOG_INF("Button short press");
                    break;
                case BUTTON_EVENT_LONG:
                    main_evt.type = MAIN_EVENT_BTN_LONG_PRESS;
                    LOG_INF("Button long press");
                    break;
                case BUTTON_EVENT_EXTRA:
                    main_evt.type = MAIN_EVENT_BTN_EXTRA;
                    LOG_INF("Button extra long press");
                    break;
                case BUTTON_EVENT_LONGEST:
                    main_evt.type = MAIN_EVENT_BTN_LONGEST;
                    LOG_INF("Button longest press");
                    break;
                default:
                    continue; // ignoruj neznámé eventy
            }
            main_evt.param = 0; // můžeš poslat index tlačítka, když budeš mít víc tlačítek

            // Pošli do Main Tasku:
            TaskMain_PutEvent(&main_evt);
        }
    }
}
