#include <zephyr/kernel.h>
#include <zephyr/drivers/gpio.h>
#include <zephyr/logging/log.h>
#include "TaskLED.h"
#include "hw_init.h"
#include "main.h"

LOG_MODULE_REGISTER(task_led, LOG_LEVEL_INF);

#define STACK_SIZE 1024
#define LED_PRIORITY 0
#define LED_MSGQ_LEN 16

typedef enum
{
    LED_STATE_IDLE,
    LED_STATE_PATTERN
} led_state_t;

typedef struct
{
    struct gpio_dt_spec gpio;
    struct k_timer timer;
    const led_pattern_t *pattern;
    uint8_t current_step;
    uint32_t repeat_counter;
    led_state_t state;
    bool has_pending_msg;
    led_msg_t pending_msg;
} led_ctx_t;

// Patterny
static const led_pattern_step_t pattern_short_press[] =
{
    { true, 100 },  { false, 1000 }
};
static const led_pattern_step_t pattern_heartbeat[] =
{
    { true, 60 },  { false, 900 }
};
static const led_pattern_step_t pattern_system_on[] =
{
    { true, 60 },{ false, 250 }, { true, 30 }, { false, 2500 }
};

static const led_pattern_t led_patterns[] =
{
    [LED_SHORT_PRESS] = { pattern_short_press, ARRAY_SIZE(pattern_short_press) },
    [LED_SYSTEM_ON] = { pattern_system_on, ARRAY_SIZE(pattern_system_on) },
    [LED_HEARTBEAT] = { pattern_heartbeat, ARRAY_SIZE(pattern_heartbeat) }
};

K_THREAD_STACK_DEFINE(led_stack_area, STACK_SIZE);
static struct k_thread led_thread;
K_MSGQ_DEFINE(led_msgq, sizeof(led_msg_t), LED_MSGQ_LEN, 4);
static led_ctx_t leds[LED_COUNT];

static void Callback_LED_Timer(struct k_timer *timer);
static void Task_LED(void *a, void *b, void *c);

// Vrací pointer na pattern, nebo NULL
static const led_pattern_t *get_pattern(led_cmd_t cmd)
{
    switch (cmd)
    {
        case LED_SHORT_PRESS:
            return &led_patterns[LED_SHORT_PRESS];
        case LED_SYSTEM_ON:
            return &led_patterns[LED_SYSTEM_ON];
        case LED_HEARTBEAT:
            return &led_patterns[LED_HEARTBEAT];
        default:
            return NULL;
    }
}

// Spuštění nebo přerušení patternu
static void led_start_pattern(led_ctx_t *led, const led_msg_t *msg)
{
    const led_pattern_t *pat = get_pattern(msg->cmd);
    led->pattern = pat;
    led->current_step = 0;
    led->repeat_counter = msg->repeat_count;
    led->state = (pat != NULL) ? LED_STATE_PATTERN : LED_STATE_IDLE;
    led->has_pending_msg = false;

    if (pat != NULL && pat->step_count > 0)
    {
        gpio_pin_set_dt(&led->gpio, pat->steps[0].on ? 1 : 0);
        k_timer_start(&led->timer, K_MSEC(pat->steps[0].duration_ms), K_NO_WAIT);
    }
    else if (msg->cmd == LED_CMD_ON)
    {
        gpio_pin_set_dt(&led->gpio, 1);
        k_timer_stop(&led->timer);
    }
    else
    {
        gpio_pin_set_dt(&led->gpio, 0);
        k_timer_stop(&led->timer);
    }
}

// Callback pro timer – posun patternu, nebo přechod na další pending zprávu
static void Callback_LED_Timer(struct k_timer *timer)
{
    led_ctx_t *led = CONTAINER_OF(timer, led_ctx_t, timer);

    if (led->pattern == NULL)
    {
        return;
    }

    led->current_step++;

    if (led->current_step >= led->pattern->step_count)
    {
        led->current_step = 0;
        if (led->repeat_counter != INFINITY_BLINK_T)
        {
            if (led->repeat_counter > 0)
            {
                led->repeat_counter--;
            }
            if (led->repeat_counter == 0)
            {
                led->state = LED_STATE_IDLE;
                gpio_pin_set_dt(&led->gpio, 0);

                if (led->has_pending_msg)
                {
                    led_start_pattern(led, &led->pending_msg);
                    led->has_pending_msg = false;
                }
                return;
            }
        }
    }
    if (led->pattern != NULL)
    {
        gpio_pin_set_dt(&led->gpio, led->pattern->steps[led->current_step].on ? 1 : 0);
        k_timer_start(&led->timer, K_MSEC(led->pattern->steps[led->current_step].duration_ms), K_NO_WAIT);
    }
}

void TaskLED_Start(void)
{
    k_thread_create(&led_thread, led_stack_area, K_THREAD_STACK_SIZEOF(led_stack_area),
        Task_LED, NULL, NULL, NULL,
        LED_PRIORITY, 0, K_NO_WAIT);

    k_thread_name_set(&led_thread, "TaskLED");
}

int TaskLED_SendCommand(const led_msg_t *msg)
{
    return k_msgq_put(&led_msgq, msg, K_FOREVER);
}

// ------- *** TASK JE POSLEDNÍ FUNKCE V SOUBORU! *** -------
static void Task_LED(void *a, void *b, void *c)
{
    leds[0].gpio = *get_led_red();
    leds[1].gpio = *get_led_green();

    for (int i = 0; i < LED_COUNT; i++)
    {
        k_timer_init(&leds[i].timer, Callback_LED_Timer, NULL);
        leds[i].state = LED_STATE_IDLE;
        leds[i].pattern = NULL;
        leds[i].current_step = 0;
        leds[i].repeat_counter = 0;
        leds[i].has_pending_msg = false;
        gpio_pin_set_dt(&leds[i].gpio, 0);
    }

    led_msg_t msg;

    while (1)
    {
        log_stack_usage("TaskLED");

        k_msgq_get(&led_msgq, &msg, K_FOREVER);

        // --- OŠETŘENÍ INDEXU LED ---
        if (msg.led_index >= LED_COUNT)
        {
            LOG_ERR("LED index %d out of range! (max %d)", msg.led_index, LED_COUNT-1);
            continue; 
        }

        led_ctx_t *led = &leds[msg.led_index];

        if (msg.wait_for_previous && led->state == LED_STATE_PATTERN)
        {
            led->pending_msg = msg;
            led->has_pending_msg = true;
        }
        else
        {
            led_start_pattern(led, &msg);
        }
    }
}
